package require xds

#
# Procedure - chech "status" variable of target-side code
# and exit with error info if error
#
proc check_error {dev stadr stpage} {

  set st [ ReadLong $dev $stadr $stpage ]

  if { $st != 0 } {
    switch $st {
      1 { set err "Internal error: Invalid command" }
      2 { set err "Unexpected NAK while communicating with I2C EEPROM" }
      3 { set err "Data verification failed" }
      4 { set err "Internal error: Flash type not supported" }
      5 { set err "Flash reports timeout"}
      6 { set err "Internal error: function not implemented"}
      7 { set err "Internal error: invalid parameter"}
      8 { set err "Flash is dirty"}
      9 { set err "Flash compacting failed"}
     10 { set err "Flash erasing failed"}
     11 { set err "Flash key mismatch error"}
     12 { set err "Flash programming error"}
     13 { set err "Not enabled attempt to program non-default flash protection keys"}
     14 { set err "Start address incorrectly aligned"}
     15 { set err "Block size incorrectly aligned"}
      default { set err "Internal error: Unexpected error code" }
    }
    xerror $dev "Error: $err" 1
  }
}

####################################################
#
# main code begins here
#
####################################################

set mypath [ file dirname $argv0 ]

#
# parse cmdline:
#
# sauflash <emu>:<board>:<cpu>:<iface>:<flash> program binary.file ?address?
# sauflash <emu>:<board>:<cpu>:<iface>:<flash> erasesector address
# sauflash <emu>:<board>:<cpu>:<iface>:<flash> erasechip
#

if { $argc < 2 } {
  puts "Flash/EEPROM programming tool for SAU510 series JTAG emulators"
  puts "Copyright (c) 2009, 2010, 2011, 2012 Sauris GmbH"
  puts ""
  puts "USAGE:"
  puts "sauflash <emu>:<board>:<cpu>:<iface>:<flash> program binary.file ?address?"
  puts "sauflash <emu>:<board>:<cpu>:<iface>:<flash> erasesector address"
  puts "sauflash <emu>:<board>:<cpu>:<iface>:<flash> erasesectors sec1,sec2,sec3-sec4,..."
  puts "sauflash <emu>:<board>:<cpu>:<iface>:<flash> erasechip"
  exit 2
}

set emu_s [lindex [split [lindex $argv 0] ':' ] 0 ]
set brd_s [lindex [split [lindex $argv 0] ':' ] 1 ]
set cpu_s [lindex [split [lindex $argv 0] ':' ] 2 ]
set if_s  [lindex [split [lindex $argv 0] ':' ] 3 ]
set ic_s  [lindex [split [lindex $argv 0] ':' ] 4 ]

if { ! [file exists "$mypath/init/emu/$emu_s.tcl"] } {
  puts "Incorrect emulator specified: $emu_s"
  puts "Emulator init script not found: $mypath/init/emu/$emu_s.tcl"
  exit 2
}

if { ! [file exists "$mypath/init/cpu/$cpu_s.tcl"] } {
  puts "Incorrect CPU specified: $cpu_s"
  puts "CPU init script not found: $mypath/init/cpu/$cpu_s.tcl"
  exit 2
}

if { ! [file exists "$mypath/init/ic/$ic_s.tcl"] } {
  puts "Incorrect Flash/EEPROM IC specified: $ic_s"
  puts "IC init script not found: $mypath/init/ic/$ic_s.tcl"
  exit 2
}

if { ! [file exists "$mypath/init/brd/$brd_s.tcl"] } {
    puts "Incorrect board specified: $brd_s"
    puts "Board init script not found: $mypath/init/brd/$brd_s.tcl"
    exit 2
}

if { ! [file exists "$mypath/init/if/$if_s.tcl"] } {
    puts "Incorrect interface specified: $if_s"
    puts "Interface init script not found: $mypath/init/if/$if_s.tcl"
    exit 2
}

puts "Initializing emulator parameters - calling $emu_s.tcl"
source $mypath/init/emu/$emu_s.tcl
puts "Initializing board parameters - calling $brd_s.tcl"
source $mypath/init/brd/$brd_s.tcl
puts "Initializing CPU parameters - calling $cpu_s.tcl"
source $mypath/init/cpu/$cpu_s.tcl
puts "Initializing interface parameters - calling $if_s.tcl"
source $mypath/init/if/$if_s.tcl
puts "Initializing flash parameters - calling $ic_s.tcl"
source $mypath/init/ic/$ic_s.tcl

if { ! [info exists dbus_width] } {
  set dbus_width 0
}

if { -1 == [lsearch -exact $compatible_interfaces $iftype] } {
  puts "Selected IC can't be connected using selected interface"
  exit 2
}

set operation [lindex $argv 1]

if { [string equal -nocase $operation "program" ] } {

  if { $argc < 3 } {
    puts "Name of file with flash image required"
    exit 2
  }

  if { $argc > 4 } {
    puts "Too many arguments"
    exit 2
  }

  set file_name [lindex $argv 2]

  if { ! [file exists $file_name] } {
    puts "File '$file_name' not found"
    exit 2
  }

  set start_addr 0

  if { $argc == 4 } {
    set start_addr [expr [lindex $argv 3]]
  }

  set fsize [file size $file_name]

  if { $fsize == 0 } {
    puts "File '$file_name' has zero size"
    exit 2
  }

  set op "program"

}

if { [string equal -nocase $operation "erasesectors" ] } {

  if { $argc < 3 } {
    puts "Sector list to erase required"
    exit 2
  }

  if { $argc > 3 } {
    puts "Too many arguments"
    exit 2
  }

  set seclist [split [lindex $argv 2] ","]
  set sector_list ""
  foreach s $seclist {
     set sector [split $s "-"]
     set slen [llength $sector]
     if { $slen == 1 } {
       incr slen
       lappend sector [lindex $sector 0]
     }
     if { $slen != 2 } {
       puts "Invalid sector in sector list"
       exit 2
     }
     set s_start [lindex $sector 0]
     set s_end [lindex $sector 1]
     if { $s_start == "" } {
        set s_start 0
     } else {
        set s_start [expr $s_start]
     }
     if { $s_end == "" } {
        set s_end 2147483647
     } else {
        set s_end [expr $s_end]
     }
     while { $s_start <= $s_end } {
       if { $uniform_sector != 0 } {
          set pages [expr $flash_size / $sectorsize]
       } else {
	  set pages [llength $sectorsize]
       }
       if { $s_start >= $pages } { break }
       lappend sector_list $s_start
       incr s_start
     }
  }

  if { [llength $sector_list] == 0 } {
     puts "No valid sectors in sector list"
     exit 2
  }

  set op "listerase"

}

if { [string equal -nocase $operation "erasesector" ] } {

  if { $argc < 3 } {
    puts "Address within sector to erase required"
    exit 2
  }

  if { $argc > 3 } {
    puts "Too many arguments"
    exit 2
  }

  set start_addr [expr [lindex $argv 2]]

  set op "secerase"

}

if { [string equal -nocase $operation "erasechip" ] } {

  if { $argc > 2 } {
    puts "Too many arguments"
    exit 2
  }

  set start_addr 0

  set op "allerase"

}


if {! [info exists op] } {
  puts "Invalid operation specified: '$operation'"
  exit 2
}

if { [string equal $op "program"] } {

  set temp [expr $fsize + $start_addr]

  if { $temp > $flash_size } {
    puts "Sum of size of '$file_name' and start address ($temp) is greater than ROM size ($flash_size)"
    exit 2
  }

  set alignment_mask [expr 7 >> (3-$dbus_width)]

  if { $start_addr & $alignment_mask } {
    puts "Start address incorrectly aligned"
    exit 2
  }

}

if { [string equal $op "secerase"] } {

  if { $start_addr >= $flash_size } {
    puts "Address is outside of a chip ($start_addr >= $flash_size)"
    exit 2
  }

}

# check that flash type is supported on this CPU

if { [ lsearch -exact $flash_types $flash_type ] < 0 } {
  puts "Flash type $flash_type is not supported on $cpu_s"
  exit 2
}

# User proc, called before connect - for prepare CPU's environment
# if it is slave proceccor, for example.

if { [llength [info procs UserCallBeforeConnect]] } {
  puts "UserCallBeforeConnect."
  UserCallBeforeConnect $DEV
}

set ii 0
while { 1 } {
  set result  [xds connect $DEV]
  if { $result == 0 } {
    puts "Connected to target"
    break
  }
  if { $ii == 4} {
    xerror $DEV "Error connecting to target" 1
  }
  incr ii
  puts "Attempt #$ii connecting to target failed:"
  puts [xds fetch_errors $DEV]
}

set result [xds reset $DEV]
if { $result != 0 } {
  xerror $DEV "Error resetting target CPU" 1
}

if { [llength [info procs UserInitAfterReset]] } {
  puts "UserInitAfterReset."
  UserInitAfterReset $DEV
}

if { [llength [info procs CPUSpecificInit]] } {
  puts "CPUSpecificInit."
  CPUSpecificInit $DEV
}

if { [llength [info procs UserInitAfterCPUInit]] } {
  puts "UserInitAfterCPUInit."
  UserInitAfterCPUInit $DEV
}

puts "Uploading [file tail $FlashLoadSuppCode] to target." 

# Load COFF file to memory, get COFF flags

set coff [xdsutil coff load $FlashLoadSuppCode]
set flags [xdsutil coff flags $coff]

# Upload COFF file into target

set result [ldpgm $coff $DEV]

if { $result != 0 } {
  xerror $DEV "Error uploading target-side code" 1
}

# Test target endianess

puts -nonewline "Testing target endianess ..."

getsym $coff _endian_test
set tmp [ ReadLong $DEV $_endian_test_addr $_endian_test_pg ]
if { $tmp != 0x12345678 } {
  puts ""
  ferror $DEV [format "Bad target endianess: ReadLong returns 0x%08x, expected 0x12345678" $tmp] 1
}

puts " ok"

# Setup breakpoint if needed
# Check COFF for "_BreakPoint" symbol presence
# If exists, set breakpoint at address, to which this symbol points

set syms [xdsutil coff symbols $coff]
if { [lsearch -exact $syms "_BreakPoint"] >= 0 } {

  set bkpt_sym [xdsutil coff symbol $coff "_BreakPoint"]

  set bkpt_addr [expr [lindex [split [lsearch -inline $bkpt_sym "VALUE*"] "="] 1]]
  set bkpt_pg [expr [lindex [split [lsearch -inline $bkpt_sym "MEMPAGE*"] "="] 1]]

  set hexad [format "%08x" $bkpt_addr]
  puts "Setting breakpoint @$bkpt_pg:$hexad"

  set result [xds setbp $DEV $bkpt_addr $bkpt_pg]
  if { $result != 0 } {
    xerror $DEV "Error setting breakpoint" 1
  }

}

# Run CPU
puts "Starting Flash loader target-side code."

# Set CPU's PC register to program's entry point

set entrypoint [expr [lindex [split [lsearch -inline $flags "ENTRY*"] "="] 1]]
set data [format "%x" $entrypoint]
set result [xds write_reg $DEV $data $pc_name]
if { $result != 0 } {
  xerror $DEV "Error setting entry point for target-side code" 1
}

set result [xds run $DEV]
if { $result != 0 } {
  xerror $DEV "Error starting target-side code" 1
}

# Wait for CPU halted due to a breakpoint in flash loader

WaitForBP $DEV 10

puts "Flash loader target-side code is now ready."

# Flash loader is now ready to accept commands/data

# Read loader buffer parameters from COFF file and target
getsym $coff _buffer
getsym $coff _buffer_size
set buffer_size [ ReadLong $DEV $_buffer_size_addr $_buffer_size_pg ]
puts [format "Loader's buffer @%u:0x%x with size of %u bytes" $_buffer_pg $_buffer_addr $buffer_size]

# Get addresses of variables from COFF file
getsym $coff _command
getsym $coff _status
getsym $coff _address
getsym $coff _length
getsym $coff _flash_type
getsym $coff _flash_size
getsym $coff _width
getsym $coff _word_width

# Setup parameters

# for all flash types
PutLong $DEV $_flash_type_addr $_flash_type_pg $flash_type
PutLong $DEV $_flash_size_addr $_flash_size_pg $flash_size
PutLong $DEV $_width_addr $_width_pg $dbus_width
PutLong $DEV $_word_width_addr $_word_width_pg $word_width

# Interface init
if { [info exists iface_init_script] } {
  puts "Calling interface specific setup script $iface_init_script"
  source $mypath/init/if/supp/$iface_init_script
}

# Flash/EEPROM specific init
if { [info exists ic_specific_setup] } {
  puts "Calling Flash/EEPROM specific setup script $ic_specific_setup"
  source $mypath/init/ic/supp/$ic_specific_setup
}

if { [llength [info procs UserInitCPUEnv]] } {
  puts "UserInitCPUEnv."
  UserInitCPUEnv $DEV
}

if { [string equal $op "program"] } {

  puts "Performing chip programming operation"

  PutLong $DEV $_address_addr $_address_pg $start_addr

  # open file, get file size
  set fp [open $file_name r]
  fconfigure $fp -translation binary

  set i 0

  while { $fsize != 0 } {

    if { $fsize > $buffer_size } {
      set blks $buffer_size  
    } else {
      set blks $fsize
    }

    set fsize [expr $fsize - $blks]
    set bindata [read $fp $blks]
    set data [xdsutil bintohexdata $bindata]

    puts "Block #$i $blks bytes"
    incr i

    set result [xds write_mem_coff $DEV $_buffer_addr $blks $_buffer_pg $data]
    if { $result != 0 } {
      xerror $DEV "Error transferring data block to target" 1
    }

    PutLong $DEV $_length_addr $_length_pg $blks
    PutLong $DEV $_command_addr $_command_pg 1

    set result [xds run $DEV]
    if { $result != 0 } {
      xerror $DEV "Error starting target-side code" 1
    }

  # calculate timeout value
  
    if { $programming_by_byte } {
      set tm [expr ( $blks + 63 ) / 64 ]
    } else {
      set cblk [expr ( $blks + $pagesize - 1 ) / $pagesize ]
      set tm [expr ($cblk + 5) ]
    }

    WaitForBP $DEV $tm

    PutLong $DEV $_command_addr $_command_pg 0
  
    check_error $DEV $_status_addr $_status_pg
  }

  close $fp
}

if { [string equal $op "secerase"] } {

  puts "Performing chip sector erase operation"

  PutLong $DEV $_address_addr $_address_pg $start_addr
  PutLong $DEV $_command_addr $_command_pg 2

  set result [xds run $DEV]
  if { $result != 0 } {
    xerror $DEV "Error starting target-side code" 1
  }

  # Fixed timeout 15 sec
  WaitForBP $DEV 150 

  PutLong $DEV $_command_addr $_command_pg 0

  check_error $DEV $_status_addr $_status_pg

}

if { [string equal $op "listerase"] } {
  puts "Performing erase operation on list of sectors"

  foreach s $sector_list {

    if { $uniform_sector != 0 } {
      set addr [expr $s * $sectorsize]
    } else {
      set addr 0
      for { set i 0 } { $i < $s } { incr i } {
        set addr [expr $addr + [lindex $sectorsize $i]]
      }
    }

    puts [format "Erasing page #%i @0x%08x" $s $addr]

    PutLong $DEV $_address_addr $_address_pg $addr
    PutLong $DEV $_command_addr $_command_pg 2

    set result [xds run $DEV]
    if { $result != 0 } {
      xerror $DEV "Error starting target-side code" 1
    }

    # Fixed timeout 15 sec
    WaitForBP $DEV 150 

    PutLong $DEV $_command_addr $_command_pg 0

    check_error $DEV $_status_addr $_status_pg
    
  }

}


if { [string equal $op "allerase"] } {

  puts -nonewline "Performing full chip page-by-page erase operation"

  if { $uniform_sector != 0 } {

    set pages [expr $flash_size / $sectorsize]

    puts " ($pages pages)"

    while { $pages != 0 } {

      set pages [expr $pages - 1]

      puts "Erasing page #$pages of $sectorsize bytes"
    
      set addr [expr $pages * $sectorsize]
      PutLong $DEV $_address_addr $_address_pg $addr
      PutLong $DEV $_command_addr $_command_pg 2

      set result [xds run $DEV]
      if { $result != 0 } {
        xerror $DEV "Error starting target-side code" 1
      }

      # Fixed timeout 15 sec
      WaitForBP $DEV 150 

      PutLong $DEV $_command_addr $_command_pg 0

      check_error $DEV $_status_addr $_status_pg

    }

  } else {
    
    set pages [llength $sectorsize]
    puts "($pages pages)"

    set addr 0
    set page 0

    foreach pg_size $sectorsize {

      puts "Erasing page #$page of $pg_size bytes"
    
      PutLong $DEV $_address_addr $_address_pg $addr
      PutLong $DEV $_command_addr $_command_pg 2

      set result [xds run $DEV]
      if { $result != 0 } {
        xerror $DEV "Error starting target-side code" 1
      }

      # Fixed timeout 15 sec
      WaitForBP $DEV 150 

      PutLong $DEV $_command_addr $_command_pg 0

      check_error $DEV $_status_addr $_status_pg
 
      set addr [expr $addr + $pg_size]
      incr page

    }

  }

}

puts "Ok."

if { [llength [info procs UserFinishCPUEnv]] } {
  puts "UserFinishCPUEnv."
  UserFinishCPUEnv $DEV
}

xds disconnect $DEV
xds quit $DEV
unload $TCLXDSLIBFULL
